program MonitorClass;

uses
  SysUtils, StrUtils, Exec, Utility, Intuition, AGraphics;

{$mode delphi}{$H+}

const
  pfnames: array[0..13] of string =
  ( 'LUT8',
    'RGB15', 'BGR15', 'RGB15PC', 'BGR15PC',
    'RGB16', 'BGR16', 'RGB16PC', 'BGR16PC',
    'RGB24', 'BGR24',
    'ARGB32', 'BGRA32', 'RGBA32'
  );

  depths: array[0..4] of Byte = (8, 15, 16, 24, 32);


procedure ReadMonitorClass;
var
  Monitors, Mon: PPObject_;
  Name, DrvName: STRPTR;
  pfs: PLongWord;
  pf: LongInt;
  i: Integer;
  gpf: TmsGetDefaultPixelFormat;
begin
  Monitors := GetMonitorList(nil);
  if Monitors = nil then
  begin
    WriteLn('No Monitors');
    Exit;
  end;
  Mon := Monitors;
  while Mon^ <> nil do
  begin
    GetAttr(MA_MonitorName, Mon^, PIPTR(@Name));
    GetAttr(MA_DriverName, Mon^, PIPTR(@DrvName));
    writeln(Format('Monitor $%p %s %s', [Mon^, Name, DrvName]));

    writeln('Supported pixelformats:');
    GetAttr(MA_PixelFormats, Mon^, PIPTR(@pfs));
    for i := Low(pfnames) to High(pfNames) do
    begin
      Writeln(Format(' %7s %s', [pfnames[i], IfThen(pfs[i] <> 0, 'yes', 'no')]));
    end;

    writeln('Preferred pixelformats:');
    gpf.MethodID := MM_GetDefaultPixelFormat;
    gpf.Store := @pf;
    for i := 0 to High(Depths) do
    begin
      write(Format(' %2d ', [depths[i]]));
      gpf.Depth := depths[i];
      DoMethodA(Mon^, @gpf);
      if (pf >= Low(pfnames)) and (pf <= High(pfnames)) then
        writeln(pfnames[pf])
      else
        if pf = -1 then
          writeln('Not Supported')
        else
          writeln(Format('Unknown (%d)', [pf]));
    end;

    writeln();
    Inc(Mon);
  end;
  FreeMonitorList(Monitors);
end;


begin
  ReadMonitorClass;
end.
